/*
 * Antonio Ronaldo Nugoli 
 * Rgm: 015724
 * Universidade Estadual de Mato  Grosso do Sul - UEMS
 * Sistemas de Informao
 * Projeto Final de Curso II
 * 
 * Implementao de um Game Interativo Utilizando a Biblioteca OpenCV
 * Jogo: Nugoli Racing Web Cam
 * 
 * Descrio da Classe:
 * Junto com a biblioteca OpenCV gerencia a webcam e o rastreamento do quadriltero atualizando a posio do jogador na tela
 */

package jogo;

import java.util.ArrayList;
import java.util.List;

import org.opencv.core.Core;
import org.opencv.core.CvType;
import org.opencv.core.Mat;
import org.opencv.core.MatOfPoint;
import org.opencv.core.MatOfPoint2f;

import org.opencv.core.Point;

import org.opencv.core.Scalar;
import org.opencv.core.Size;
import org.opencv.highgui.Highgui;
import org.opencv.highgui.VideoCapture;
import org.opencv.imgproc.Imgproc;

import telas.Camera;

public class Webcam implements Runnable {

	private VideoCapture capture = null;

	// Guarda a imagem capturada pela cmera
	private Mat webcan_image = new Mat();
	private Mat webcan_image2 = new Mat();

	private Camera camera;
	private Jogador jogador;
	private MatParaBuff matbuff;
	private boolean liga = true;
	private int larguraTela;
	private RenderPista pista;
	private int decrementoy, incrementoy;

	public Webcam(Jogador jogador, Camera camera, RenderPista pista) throws InterruptedException {

		this.pista = pista;
		this.jogador = jogador;
		this.camera = camera;
		larguraTela = jogador.getLargura();

		// Define o espao que jogador move a cada interao em relao a
		// resoluo de tela
		switch (larguraTela) {
		case 600:
			decrementoy = -5;
			incrementoy = 5;
			break;

		case 800:
			decrementoy = -7;
			incrementoy = 7;
			break;

		case 1024:
			decrementoy = -8;
			incrementoy = 8;
			break;
		}

		matbuff = new MatParaBuff();

		// Inicia a webcam e configura o tamanho da imagem capturada
		capture = new VideoCapture(0);
		if (capture.isOpened()) {
			capture.set(Highgui.CV_CAP_PROP_FRAME_WIDTH, 300);
			capture.set(Highgui.CV_CAP_PROP_FRAME_HEIGHT, 300);
		}
	}

	public void run() {
		while (liga) {

			// Evita que o carro do jogador sai da pista
			if ((jogador.getX() < ((larguraTela * 12) / 100)))
				jogador.setX((larguraTela * 12) / 100);
			if ((jogador.getX() > ((larguraTela * 66) / 100)))
				jogador.setX((larguraTela * 66) / 100);

			// Captura a imagem da webcam
			capture.read(webcan_image);

			// Espelha a imahem
			Core.flip(webcan_image.clone(), webcan_image, 1);

			// Cria uma cpia da imagem
			webcan_image2 = webcan_image.clone();

			// Transforma a imagem para escala de cinza
			Imgproc.cvtColor(webcan_image2.clone(), webcan_image2, Imgproc.COLOR_BGR2GRAY);

			// Ajusta e binariza a imagem os valores de corte de Threshold foram
			// ajustadas para retngulos escuros sobre superfcies claras
			Imgproc.GaussianBlur(webcan_image2.clone(), webcan_image2, new Size(5, 5), 5);
			Imgproc.adaptiveThreshold(webcan_image2.clone(), webcan_image2, 240, 1, 1, 85, 10);

			List<MatOfPoint> contours = new ArrayList<MatOfPoint>();

			// Busca contornos na imagem binarizada
			Imgproc.findContours(webcan_image2.clone(), contours, new Mat(), Imgproc.RETR_LIST,
					Imgproc.CHAIN_APPROX_SIMPLE);

			// Para cada contorno analisa em busca de contornos interligados
			for (int i = 0; i < contours.size(); i++) {

				MatOfPoint2f card2f = new MatOfPoint2f(contours.get(i).toArray());
				if (Imgproc.contourArea(contours.get(i)) > 100) {

					double peri = Imgproc.arcLength(card2f, true);
					MatOfPoint2f approx = new MatOfPoint2f();

					MatOfPoint approxf1 = new MatOfPoint();
					Imgproc.approxPolyDP(card2f, approx, 0.02 * peri, true);

					approx.convertTo(approxf1, CvType.CV_32S);
					List<MatOfPoint> contourTemp = new ArrayList<>();
					contourTemp.add(approxf1);
					int cont = 1;
					double x1 = 0, x2 = 0, y1 = 0, y2 = 0, pontosX[] = { 0, 0, 0, 0 }, pontosY[] = { 0, 0, 0, 0 };
					boolean pintar = false;

					// Seleciona apenas quadrilateros
					if (contourTemp.get(0).rows() == 4) {

						for (MatOfPoint mop : contourTemp) {
							for (Point p : mop.toList()) {

								if (cont == 1) {
									x1 = p.x;
									y1 = p.y;
								}
								if (cont == 3) {
									x2 = p.x;
									y2 = p.y;
								}
								pontosX[cont - 1] = p.x;
								pontosY[cont - 1] = p.y;
								cont++;
							}

							// Tenta refinar o rastreamento excluido objetos
							// localizados nas extremidades da imagem ou muito
							// pequenos
							if (((x1 + x2) / 2 > 50 && (x1 + x2) / 2 < 200)
									&& ((y1 + y2) / 2 > 50 && (y1 + y2) / 2 < 200)) {

								if (calculaAreaRetangulo(pontosX, pontosY)) {
									pista.setIniciou(true);
									pintar = true;
								}

							}

							// Caso o rastreamento retorne um quadriltero
							// comea a desenhar na imagem
							if (pintar) {
								jogador.setDx(0);
								Core.fillPoly(webcan_image, contourTemp, new Scalar(255, 0, 0));
								Core.circle(webcan_image, new Point(pontosX[0], pontosY[0]), 4, new Scalar(0, 0, 0), 2);
								Core.circle(webcan_image, new Point(pontosX[1], pontosY[1]), 4, new Scalar(0, 255, 0),
										2);
								Core.circle(webcan_image, new Point(pontosX[2], pontosY[2]), 4,
										new Scalar(255, 255, 255), 2);
								Core.circle(webcan_image, new Point(pontosX[3], pontosY[3]), 4, new Scalar(255, 0, 255),
										2);

								Core.line(webcan_image, new Point(((x1 + x2) / 2) - 100, (y1 + y2) / 2),
										new Point(((x1 + x2) / 2) + 100, (y1 + y2) / 2), new Scalar(255, 255, 0));

								Core.circle(webcan_image, new Point((x1 + x2) / 2, ((((y2 - y1) * 20) / 100)) + y1), 2,
										new Scalar(255, 255, 255), 2);

								/* Linha de zona morta */
								Core.line(webcan_image,
										new Point(((x1 + x2) / 2) - 100, ((((y2 - y1) * 20) / 100)) + y1),
										new Point(((x1 + x2) / 2) + 100, ((((y2 - y1) * 20) / 100)) + y1),
										new Scalar(255, 255, 0));

								Core.circle(webcan_image, new Point((x1 + x2) / 2, (y1 + y2) / 2), 4,
										new Scalar(255, 0, 255), 4);

								System.out.println((pontosX[3] - pontosX[0]) + " " + (pontosX[0] - pontosX[1]));

								// Fixa o ponto de zona morta
								double centery = ((((y2 - y1) * 20) / 100)) + y1;

								// Verifica se um dos pontos superiores do
								// quadriltero ficou abaixo da linha de zona
								// morta para virar o veculo
								if ((pontosX[3] - pontosX[0]) > (pontosX[0] - pontosX[1])) {
									if (pontosY[3] > centery) {
										System.out.println("curva a direita 1");

										if (jogador.getX() < ((larguraTela * 66) / 100))// enguando
																						// ocupar
																						// 66%
																						// do
																						// total
																						// da
																						// tela
											jogador.setDx(incrementoy);
										else {
											jogador.setDx(0);
										}
									}
									if (pontosY[0] > centery) {
										System.out.println("curva a esquerda 2");
										if (jogador.getX() > ((larguraTela * 12) / 100)) // enguando
																							// no
																							// ocupar
																							// 12%
																							// do
																							// inicio
																							// da
																							// tela
											jogador.setDx(decrementoy);
										else {
											jogador.setDx(0);
										}
									}

								}

								else {

									if (pontosY[0] > centery) {

										// Controla o deslocamento do jogador
										// para no sair fora da viso da tela
										if (jogador.getX() < ((larguraTela * 66) / 100)) // enguando
																							// ocupar
																							// 66%
																							// do
																							// total
																							// da
																							// tela
											jogador.setDx(incrementoy);
										else {
											jogador.setDx(0);
										}

									}
									if (pontosY[1] > centery) {
										if (jogador.getX() > ((larguraTela * 12) / 100)) // enguando
																							// no
																							// ocupar
																							// 12%
																							// do
																							// inicio
																							// da
																							// tela
											jogador.setDx(decrementoy);
										else {
											jogador.setDx(0);
										}
									}

								}

							}

							pintar = false;
							cont = 1;

						}
					}
				}
			}

			try {
				matbuff.setMatrix(webcan_image, ".jpg");

				// envia para camera a imagem da webcam para ser visualizada na
				// tela
				camera.setImage(matbuff.getBufferedImage());
			} catch (Exception e) {
			}

		}
		capture.release();
	}

	// Tenta refinar a deteco eliminando objetos muito pequenos
	private boolean calculaAreaRetangulo(double pontosX[], double pontosY[]) {
		boolean resultado = false;
		double h, b1, b2, base;

		if ((pontosX[3] - pontosX[0]) > (pontosX[0] - pontosX[1])) {
			h = pontosY[1] - pontosY[0];
			b1 = pontosX[3] - pontosX[0];
			b2 = pontosX[2] - pontosX[1];
		}

		else {
			h = pontosY[2] - pontosY[1];
			b1 = pontosX[0] - pontosX[1];
			b2 = pontosX[3] - pontosX[2];
		}

		base = ((b1 + b2) / 2) * h;
		if (base > 200)
			resultado = true;
		return resultado;
	}

	public void finalisa() {
		liga = false;
	}

	public void setLiga(boolean liga) {
		this.liga = liga;
	}
}
