/*
 * Antonio Ronaldo Nugoli 
 * Rgm: 015724
 * Universidade Estadual de Mato  Grosso do Sul - UEMS
 * Sistemas de Informao
 * Projeto Final de Curso II
 * 
 * Implementao de um Game Interativo Utilizando a Biblioteca OpenCV
 * Jogo: Nugoli Racing Web Cam
 * 
 * Descrio da Classe:
 * Exibi a janela onde a partida acontece
 * 
 */

package telas;

import java.awt.Canvas;
import java.awt.Cursor;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Point;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;
import java.awt.image.AffineTransformOp;
import java.awt.image.BufferStrategy;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.util.Random;
import javax.imageio.ImageIO;
import javax.swing.ImageIcon;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import org.opencv.core.Core;
import audio.Audio;
import jogo.Jogador;
import jogo.Pista;
import jogo.RenderPista;
import jogo.Webcam;
import placar.RegistraPlacar;

public class Fase extends Canvas implements Runnable {

	private static final long serialVersionUID = 1L;
	private Thread thread, threadCamera;
	private boolean running = false;
	private TelaPrincipal tela;
	private JFrame frame;
	private Pista[] pista;
	private RenderPista desenhaPista;
	private Jogador jogador;
	private Camera camera = new Camera();
	private BufferedImage arvore, arvore1, arvore2, arvore3, arvore4, arvore5, largada, volante;
	private BufferStrategy bs;

	private Webcam webcam;
	private Audio som, somFundo;
	private File musica1, musica2;

	// Array para guardar os sprites dos carros e placas que vemos na pista;
	private BufferedImage carros[], placas[];

	// Vetor com valores utilizados para posicionar itens na pista de forma que
	// sua posio seja escolhida aleatoriamente
	private float direcao[] = { (float) -0.05, (float) -0.99 }, direcao2[] = { (float) -2.0, (float) 1.6 };

	 //int x = 10, y = 10;

	// Variaveis de controle para pista
	private int segL = 200;// altura de cada segmento da pista
	private int distanciaTotal = 6000; // Comprimento total da pista

	// Cria a janela de exibio da partida usando a resoluo desejada
	public Fase(int largura, int altura) throws InterruptedException {

		frame = new JFrame("Nugoli Racing Web Cam");
		frame.setSize(largura, altura);
		frame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
		frame.setResizable(false);
		frame.setLocationRelativeTo(null);
		frame.add(this);
		frame.setVisible(true);
		frame.pack();
		setCursor(new Cursor(8));

		// Inicializa a pista
		carregaPista();

		jogador = new Jogador(largura, altura);
		desenhaPista = new RenderPista(pista, distanciaTotal, largura, altura, segL, jogador);
		frame.addKeyListener(new Teclado());
		System.loadLibrary(Core.NATIVE_LIBRARY_NAME);
		desenhaPista.setIniciou(false);
		webcam = new Webcam(jogador, camera, desenhaPista);
		atualizarCursor();
		// Random gerador = new Random();

		som = new Audio();
		somFundo = new Audio();
		musica2 = new File("audio\\1.wav");
		musica1 = new File("audio\\bateu.wav");
		somFundo.tocaraudio(musica2);

		// Verifica se a tela de Jogo foi finalizada pelo jogador antes do fim
		// da partida
		frame.addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent evt) {
				tela.setVisible(true);
				running = false;
				somFundo.parar();
				webcam.finalisa();
				stop();
			}
		});

		// Inicia o loop do Jogo
		start();
	}

	// Esconde o cursor do mouse para no ficar sobre a tela do jogo
	private void atualizarCursor() {
		ImageIcon referencia = new ImageIcon("img\\cursor.png");
		Image icon = referencia.getImage();
		Point hotSpot = new Point(0, 25);
		Cursor cursor = new Cursor(1);
		cursor = getToolkit().createCustomCursor(icon, hotSpot, "Cursor");
		setCursor(cursor);
	}

	// Carrega todos os dados para vetor pista e os itens como carros
	// adversrios e placas que iro aparecer aleatoriamente
	private void carregaPista() {

		// Inicializa um novo vetor de objeto pista
		pista = new Pista[distanciaTotal];

		carros = new BufferedImage[6];
		placas = new BufferedImage[5];

		// carrega todas imagens que sero utilizadas as imagens carregadas em
		// vetor so para terem seus ndices escolhidos aleatoriamente
		try {
			for (int i = 1; i < 7; i++) {
				carros[i - 1] = ImageIO.read(new File("img\\carros\\carro" + i + ".png"));
				if (i < 6)
					placas[i - 1] = ImageIO.read(new File("img\\placa" + i + ".png"));
			}

			arvore = ImageIO.read(new File("img\\5.png"));
			arvore1 = ImageIO.read(new File("img\\6.png"));
			arvore2 = ImageIO.read(new File("img\\75.png"));
			arvore3 = ImageIO.read(new File("img\\1.png"));
			arvore4 = ImageIO.read(new File("img\\2.png"));
			arvore5 = ImageIO.read(new File("img\\4.png"));
			largada = ImageIO.read(new File("img\\largada.png"));
			volante = ImageIO.read(new File("img\\volante.png"));

		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		Random random = new Random(); // Gerar nmeros aleatrios

		// posiciona os itens no vetor pista em posioes variadas
		for (int i = 0; i < distanciaTotal; i++) {

			pista[i] = new Pista(getWidth(), getHeight());
			pista[i].setZ(i * segL);

			if (i % 225 == 0) {
				pista[i].setImg_cenariox((float) -1.7);
				pista[i].setImg_cenario(arvore);
			}
			if (i % 330 == 0 && i < 3000) {
				pista[i].setImg_cenariox((float) 0.80);
				pista[i].setImg_cenario(arvore1);
			}

			if (i % 255 == 0) {
				pista[i].setImg_cenariox((float) 1.2);
				pista[i].setImg_cenario(arvore4);
			}

			if (i % 280 == 0 && i < 3000) {
				pista[i].setImg_cenariox((float) -0.80);
				pista[i].setImg_cenario(arvore5);
			}

			if (i % 88 == 0) {
				pista[i].setImg_cenariox((float) -2);
				pista[i].setImg_cenario(arvore3);
			}
			if (i == 300) {
				pista[i].setImg_cenariox((float) 1.6);
				pista[i].setImg_cenario(arvore2);
			}

			if (i % 630 == 0) {
				pista[i].setImg_cenariox((float) direcao2[random.nextInt(2)]); // -0.99
				pista[i].setImg_cenario(placas[random.nextInt(5)]);
			}

			if (i == 100) {
				pista[i].setImg_cenariox((float) -0.5); // 0.005
				pista[i].setImg_cenario(largada);
			}

			if (i % 180 == 0 && i > 200 && i < 1500) {
				pista[i].setCarro(true);
				pista[i].setImg_cenariox((float) direcao[random.nextInt(2)]); // -0.99
				pista[i].setImg_cenario(carros[random.nextInt(6)]);
			}
			if (i % 82 == 0 && i > 1501 && i < 2000) {
				pista[i].setCarro(true);
				pista[i].setImg_cenariox((float) direcao[random.nextInt(2)]); // -0.99
				pista[i].setImg_cenario(carros[random.nextInt(6)]);
			}
			if (i % 40 == 0 && i > 2001) {
				pista[i].setCarro(true);
				pista[i].setImg_cenariox((float) direcao[random.nextInt(2)]); // -0.99
				pista[i].setImg_cenario(carros[random.nextInt(6)]);
			}

			/*
			 * Define os trechos onde devem haver curvas , valores positivos
			 * curva a esquerda e negativos curvas a direita , o valor tambem
			 * indica o grau de abertura da pista aumentando o valor a curva
			 * sera mais fechada .
			 */
			if (i > 300 && i < 700)
				pista[i].setCurva((float) 0.5);
			if (i > 1000 && i < 1300)
				pista[i].setCurva((float) -1.5);

			if (i > 1400 && i < 1900)
				pista[i].setCurva((float) -1.5);
			if (i > 2700 && i < 3500)
				pista[i].setCurva((float) 1.0);

			/*
			 * Define os trechos onde devem haver ondulacoes na pista ,
			 * utilizando o calculo do seno de valor constante sobre a posicao
			 * do segmento para definir a escala do eixo y .
			 */
			
			if (i > 3 && i < 2999 || i > 4001)
				pista[i].sety((float) (Math.sin(i / 30.0) * 200));
			if (i > 3000 && i < 4000)
				pista[i].sety((float) (Math.sin(i / 30.0) * 0));

		}
	}

	// Recebe dados do frame principal
	public void enviaFrame(TelaPrincipal tela1) {
		this.tela = tela1;
	}

	//Seta as variaveis para iniciar as threads e o loop
	private synchronized void start() {
		if (running)
			return;

		running = true;
		thread = new Thread(this);
		threadCamera = new Thread(webcam);
		threadCamera.start();
		thread.start();
	}
	
	//Para a execuo do loop
	private synchronized void stop() {
		if (!running)
			return;
		running = false;
		try {
			threadCamera.join();
			thread.join();
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}

	//Inicia o loop do jogo tem como regra controlar o tempo de execuo de cara ciclo para comparar com o anterior buscando manter uma taxa fixa de FPS
	public void run() {
		long tempoanterior = System.nanoTime();
		final double fps = 60.0;
		double nanosegundos = 1000000000 / fps;
		double controle = 0;
		int updates = 0;
		int frames = 0;
		long timer = System.currentTimeMillis();
		while (running) {
			long agora = System.nanoTime();

			if (controle >= 1) {
				atualiza();
				updates++;
				controle--;
			}
			desenha();
			frames++;

			controle += (agora - tempoanterior) / nanosegundos;
			tempoanterior = agora;

			if (System.currentTimeMillis() - timer > 1000) {
				timer += 1000;
				System.out.println(updates + " Quadros, Fps " + frames);
				updates = 0;
				frames = 0;
			}
		}
		stop();

	}
	//Mantem os dados da pista e audio atualizados durante a partida
	private void atualiza() {
		desenhaPista.atualiza();
		jogador.atualiza();
		if (!somFundo.clip.isActive()) {
			somFundo.tocaraudio(musica2);
		}
	}

	
	//Agrupa todos os dados da webcam e pista para desenhar na tela
	private void desenha() {

		bs = this.getBufferStrategy();

		if (bs == null) {
			createBufferStrategy(2);
			return;
		}

		Graphics g = bs.getDrawGraphics();
		g.clearRect(0, 0, getWidth(), getHeight());

		// Verifica se o jogo acabou 
		if (desenhaPista.isFimDeJogo()) {
			webcam.finalisa();
			somFundo.parar();
			som.tocaraudio(musica1);
			ImageIcon fotoJogador = new ImageIcon("img\\imgJogador.png");
			JLabel jogador = new JLabel();
			jogador.setIcon(fotoJogador);

			RegistraPlacar arqPlacar = new RegistraPlacar();
			arqPlacar.escrverArquivo("" + (desenhaPista.getStartPos() - 1));
			String msgBateRecorde;

			msgBateRecorde = arqPlacar.getMaior() > (desenhaPista.getStartPos() - 1) ? "\n"
					: "Parabns Voce atingiu um novo recorde!\n";

			JOptionPane
					.showMessageDialog(null,
							"Voc andou: " + (desenhaPista.getStartPos() - 1 + " metros \n" + msgBateRecorde
									+ " Recorde atual: " + arqPlacar.getMaior() + " metros"),
							"Fim de Jogo", 0, fotoJogador);

			tela.setVisible(true);
			running = false;

			g.dispose();
			frame.setVisible(false);

		} else {// Caso no tenha acabado pega e desenha a imagem da camera e das posies atuais do jogador e da pista 
			g = desenhaPista.desenha(g);
			g.drawImage(jogador.getJogador(), jogador.getX(), jogador.getY() - 10, jogador.getLarg(), jogador.getAlt(),
					this);
			g.drawRect(jogador.getX() + ((jogador.getLarg() * 20 / 100) / 2), jogador.getY() + 20,
					jogador.getLarg() - (jogador.getLarg() * 20 / 100), jogador.getAlt());

			g.drawImage((Image) camera.getImage(), getWidth() / 2 - 50, 10, 150, 120, null);

			g.drawImage(rotateImage(volante, jogador.getX() / 2, (getHeight() * 25) / 100), 30,
					(getHeight() / 2) + ((getHeight() * 25) / 100) / 2, (getHeight() * 25) / 100,
					(getHeight() * 25) / 100, this);

		}

		//Mostra a imagem montada na tela
		bs.show();
		g.dispose();

	}

	//Caso necessrio possibilita mover o veiculo com as setas do teclado
	private class Teclado implements KeyListener {
		@Override
		public void keyPressed(KeyEvent e) {
			// TODO Auto-generated method stub
			jogador.keyPressed(e);
		}

		@Override
		public void keyReleased(KeyEvent e) {
			// TODO Auto-generated method stub
			jogador.keyReleased(e);

		}

		@Override
		public void keyTyped(KeyEvent e) {
			// TODO Auto-generated method stub

		}

	}

	//Possibilita que outra classe solicite a interrupo do jogo
	public void setRunning(boolean running) {
		this.running = running;
	}

	/*
	 * Mtodo responsavel em rotacionar a imagem usando seu ponto central como
	 * eixo de rotao adaptado do codigo disponivel no site
	 * http://pt.stackoverflow.com/questions/3777/rota%C3%A7%C3%A3o-de-imagem-em
	 * -java
	 * 
	 * Apenas utilizada para que o volante na parte inferior esquerda da tela rotacionasse corretamente ao ser detectado a rotao do quadriltero detectado
	 * 
	 */
	public static BufferedImage rotateImage(BufferedImage rotateImage, double angle, int dimensao) {
		AffineTransform tx = new AffineTransform();
		tx.rotate(Math.toRadians(angle), rotateImage.getWidth() / 2, rotateImage.getHeight() / 2);

		// Rotaciona as coordenadas dos cantos da imagem
		Point2D[] aCorners = new Point2D[4];

		aCorners[0] = tx.transform(new Point2D.Double(dimensao / 2, dimensao / 2), null);
		aCorners[1] = tx.transform(new Point2D.Double(dimensao, dimensao / 2), null);
		aCorners[2] = tx.transform(new Point2D.Double(dimensao / 2, dimensao), null);
		aCorners[3] = tx.transform(new Point2D.Double(dimensao, dimensao), null);

		// Obtm o valor de translao para cada eixo com um canto "escondido"
		double dTransX = 0;
		double dTransY = 0;
		for (int i = 0; i < 4; i++) {
			if (aCorners[i].getX() < 0 && aCorners[i].getX() < dTransX)
				dTransX = aCorners[i].getX();
			if (aCorners[i].getY() < 0 && aCorners[i].getY() < dTransY)
				dTransY = aCorners[i].getY();
		}

		// Aplica a translao para evitar cortes na imagem
		AffineTransform translationTransform = new AffineTransform();
		translationTransform.translate(-dTransX, -dTransY);
		tx.preConcatenate(translationTransform);

		return new AffineTransformOp(tx, AffineTransformOp.TYPE_BILINEAR).filter(rotateImage, null);
	}
}
